module Page0 = struct
  type options = Options.page

  type t = {
    uri : string;
    title : string;
    history : string list;
    options : options;
    id : LablgtkReact.id;
    parent_id : LablgtkReact.id option;
    request : [ `None | `Close | `Popup ];
  }

  type init = (LablgtkReact.id * LablgtkReact.id option * options)

  type action_common = [
    | `Uri of string
    | `Title of string
    | `SetOptions of options
    | `Close
  ]

  type action_functional = [ action_common | `RequestClose | `Popup | `PopupDone ]

  type action_ui = [ action_common | `None ]

  let create (id, parent_id, options) =
    {
      uri = "about:blank";
      title = "(empty)";
      history = [];
      options = options;
      id = id;
      parent_id = parent_id;
      request = `None;
    }

  let rewrite ~t uri = (* XXX: ***BAD*** but should be good enough for now *)
    if try (String.index uri ':' < 8) with Not_found _ -> false then
      uri
    else
      String.concat "://" [ t.options.Options.default_uri_scheme; uri ]

  let on_action_func t = function
    | `Uri uri ->
        let uri = rewrite ~t uri in
        { t with uri; history = uri :: t.history }, `Uri uri
    | `Title title -> { t with title }, `Title title
    | `SetOptions options as msg -> { t with options }, msg
    | `RequestClose -> { t with request = `Close }, `None
    | `Close as msg -> t, msg
    | `Popup -> { t with request = `Popup }, `None
    | `PopupDone -> { t with request = `None }, `None
end

module Navigation0 = struct
  module Z = Zipper
  type t = {
    z : Page0.t Z.t;
    options : Options.navigation;
  }

  type init = Options.navigation

  type action_common = [
    | `Nth of int
    | `MoveTo of int
    | `SetOptions of Options.navigation
  ]

  type action_functional = [
    action_common
    | `CloseCurrent
    | `Set of Page0.t
  ]

  type action_ui = [
    action_common
    | `Close of (LablgtkReact.id * int option)
    | `Popup of LablgtkReact.id
    | `None
  ]

  let create options =
    { z = Z.empty; options = options }

  let close_current n =
    let z = Z.remove n.options.Options.post_close_activate n.z in
    { n with z }, `Close ((Z.get n.z).Page0.id, Some (Z.pos z))

  let on_action_func n (action : action_functional) = match action with
  | `Nth i ->
      if i = Z.pos n.z then
        n, `None
      else
        { n with z = Z.nth n.z i }, `Nth i
  | `MoveTo i ->
      let cur = Z.pos n.z in
      if i = cur then
        n, `None
      else
        { n with z = Z.move n.z cur i }, `MoveTo i
  | `CloseCurrent ->
      close_current n
  | `Set e ->
      (try
        let z = Z.find (fun w -> (Z.get w).Page0.id = e.Page0.id) n.z in
        let n_of z = { n with z = Z.nth z (Z.pos n.z)  }in
        match e.Page0.request with
        | `None -> n_of (Z.set z e), `None
        | `Popup -> n_of (Z.set z e), `None
        | `Close when Z.pos z = Z.pos n.z -> close_current n
        | `Close -> n_of (Z.remove `After z), `Close (e.Page0.id, None)
      with Not_found ->
        { n with z = Z.add n.options.Options.new_tab_placement n.z e }, `None)
  | `SetOptions options as msg -> { n with options }, msg
end

module Browser0 = struct
  type options = Options.browser

  type t = {
    options : options;
    navigation : Navigation0.t;
  }

  type init = options

  type action_common = [ `SetOptions of options ]

  type action_functional = [
    action_common
    | `Set of Navigation0.t
  ]

  type action_ui = [ action_common | `None ]

  let create options =
    let navigation = Navigation0.create options.Options.navigation in
    { options; navigation }

  let save_options options =
    try
      let oc = open_out_bin "caravel.conf" in
      Marshal.to_channel oc options [];
      close_out oc
    with _ -> 
      ()

  let on_action_func t = function
    | `Set navigation -> { t with navigation }, `None
    | `SetOptions options as msg ->
        save_options options;
        { t with options }, msg
end

